# -*- coding: utf-8 -*-
#
# Copyright (c) 2013-2021 Virtual Cable S.L.U.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
#    * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright notice,
#      this list of conditions and the following disclaimer in the documentation
#      and/or other materials provided with the distribution.
#    * Neither the name of Virtual Cable S.L. nor the names of its contributors
#      may be used to endorse or promote products derived from this software
#      without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


"""
@author: Adolfo Gómez, dkmaster at dkmon dot com
"""
import datetime
import time
import logging
import typing

from uds.core.managers.stats import StatsManager, REVERSE_FLDS_EQUIV
from uds.models import Provider, Service, ServicePool, Authenticator, OSManager

logger = logging.getLogger(__name__)

# EventTupleType = typing.Tuple[datetime.datetime, str, str, str, str, int]

if typing.TYPE_CHECKING:
    from django.db import models

# Posible events, note that not all are used by every possible owner type
(
    # Login - logout
    ET_LOGIN,
    ET_LOGOUT,
    # Service access
    ET_ACCESS,
    # Cache performance
    ET_CACHE_HIT,
    ET_CACHE_MISS,
    # Platforms detected
    ET_PLATFORM,
    # Tunnel
    ET_TUNNEL_OPEN,
    ET_TUNNEL_CLOSE,
    # Os Manager
    ET_OSMANAGER_INIT,
    ET_OSMANAGER_READY,
    ET_OSMANAGER_RELEASE
) = range(11)

# Events names
EVENT_NAMES: typing.Mapping[int, str] = {
    ET_LOGIN: 'Login',
    ET_LOGOUT: 'Logout',
    ET_ACCESS: 'Access',
    ET_CACHE_HIT: 'Cache hit',
    ET_CACHE_MISS: 'Cache miss',
    ET_PLATFORM: 'Platform',
    ET_TUNNEL_OPEN: 'Tunnel open',
    ET_TUNNEL_CLOSE: 'Tunnel close',
    ET_OSMANAGER_INIT: 'OS Manager init',
    ET_OSMANAGER_READY: 'OS Manager ready',
    ET_OSMANAGER_RELEASE: 'OS Manager release',
}

(
    OT_PROVIDER,
    OT_SERVICE,
    OT_DEPLOYED,
    OT_AUTHENTICATOR,
    OT_OSMANAGER
) = range(5)

TYPES_NAMES: typing.Mapping[int, str] = {
    OT_PROVIDER: 'Provider',
    OT_SERVICE: 'Service',
    OT_DEPLOYED: 'Deployed',
    OT_AUTHENTICATOR: 'Authenticator',
    OT_OSMANAGER: 'OS Manager'
}

MODEL_TO_EVENT: typing.Mapping[typing.Type['models.Model'], int] = {
    ServicePool: OT_DEPLOYED,
    Service: OT_SERVICE,
    Provider: OT_PROVIDER,
    Authenticator: OT_AUTHENTICATOR,
    OSManager: OT_OSMANAGER
}

# Events data (fld1, fld2, fld3, fld4):
# ET_LOGIN --> on Authenticator
#     (username, srcip)
#   Note: Generated on user login on UDS web
#
# ET_PLATFORM --> on Authenticator
#     (platform, browser, version)
#   Note: Generated on user login on UDS web
#
# ET_LOGOUT --> on Authenticator
#     (username, srcip)
#   Note: Generated on user logout on UDS web. Not generated if browser is closed or session discards.
#
# ET_CACHE_HIT --> On UserService
#     (usableServicesInCacheL1,)
#   Note: Generated on assigning from cache to user
#
# ET_CACHE_MISS --> On UserService
#     (preparingServicesInCacheL1,)
#   Note: Generated on missed assigning from cache to user
#
# ET_ACCESS -> On ServicePool
#     (username, srcpi, dstip, userService_uuid)
#   Note: Generated on user access to service in UDS (that is, clicked on an service)
#
# ET_LOGIN -> On ServicePool
#     (username, knownUserIp, serviceIp, fullUserName)
#   Note: Generated by OsManager (that is, services without os manager will not trigger this)
#
# ET_LOGOUT -> On ServicePool
#     (username, knownUserIp, serviceIp, fullUserName)
#   Note: Generated by OsManager (that is, services without os manager will not trigger this)
#
# OT_TUNNEL_OPEN: -> On ServicePool
#     (username, srcip, dstip, tunnel_id)
#   Note: For HTML5, scrip = "source" string indicating tunnel type (HTML5-{RDP,RDP,VNC})
#
# OT_TUNNEL_CLOSE: -> On ServicePool
#     (duration, sent, received, tunnel_id)
#
# OT_OSMANAGER_INIT: -> On OsManager
#     (servicepool_uuid, srcip, userservice_uuid)
#
# OT_OSMANAGER_READY: -> On OsManager
#     (servicepool_uuid, srcip, userservice_uuid)
#
# OT_OSMANAGER_RELEASE: -> On OsManager
#     (servicepool_uuid, '', userservice_uuid)

# Helpers
# get owner by type and id
def getOwner(ownerType: int, ownerId: int) -> typing.Optional['models.Model']:
    if ownerType == OT_PROVIDER:
        return Provider.objects.get(pk=ownerId)
    elif ownerType == OT_SERVICE:
        return Service.objects.get(pk=ownerId)
    elif ownerType == OT_DEPLOYED:
        return ServicePool.objects.get(pk=ownerId)
    elif ownerType == OT_AUTHENTICATOR:
        return Authenticator.objects.get(pk=ownerId)
    elif ownerType == OT_OSMANAGER:
        return OSManager.objects.get(pk=ownerId)
    else:
        return None

class EventTupleType(typing.NamedTuple):
    stamp: datetime.datetime
    fld1: str
    fld2: str
    fld3: str
    fld4: str
    event_type: int

    # aliases for fields
    def __getitem__(self, item) -> typing.Any:
        if item in REVERSE_FLDS_EQUIV:
            item = REVERSE_FLDS_EQUIV[item]
        return self.__getattribute__(item)

    # Obtains the Event as a string
    def __str__(self) -> str:
        # Convert Event type to string first
        eventName = EVENT_NAMES[self.event_type]
        return '{} {} {} {} {} {}'.format(self.stamp, eventName, self.fld1, self.fld2, self.fld3, self.fld4 )


EventClass = typing.Union[Provider, Service, ServicePool, Authenticator]



def addEvent(obj: EventClass, eventType: int, **kwargs) -> bool:
    """
    Adds a event stat to specified object

    Although any counter type can be added to any object, there is a relation that must be observed
    or, otherway, the stats will not be recoverable at all:


    note: Runtime checks are done so if we try to insert an unssuported stat, this won't be inserted and it will be logged
    """

    return StatsManager.manager().addEvent(
        MODEL_TO_EVENT[type(obj)], obj.id, eventType, **kwargs
    )


def getEvents(
    obj: EventClass, eventType: int, **kwargs
) -> typing.Generator[EventTupleType, None, None]:
    """
    Get events

    Args:
        obj: Obj for which to recover stats counters
        counterType: type of counter to recover
        since: (optional, defaults to 'Since beginning') Start date for counters to recover
        to: (optional, defaults to 'Until end') En date for counter to recover
        limit: (optional, defaults to 1000) Number of counter to recover. This is an 'At most' advice. The returned number of value
               can be lower, or even 1 more than requested due to a division for retrieving object at database
        all: (optinal), indicates that get all counters for the type of obj passed in, not only for that obj.

    Returns:
        A generator, that contains pairs of (stamp, value) tuples
    """
    objType = type(obj)

    if kwargs.get('all', False):
        owner_id = None
    else:
        owner_id = obj.pk

    for i in StatsManager.manager().getEvents(
        MODEL_TO_EVENT[objType], eventType, owner_id=owner_id, since=kwargs.get('since'), to=kwargs.get('to')
    ):
        yield EventTupleType(
            datetime.datetime.fromtimestamp(i.stamp),
            i.fld1,
            i.fld2,
            i.fld3,
            i.fld4,
            i.event_type,
        )
 
 # tail the events table
def tailEvents(sleepTime: int = 2) -> typing.Generator[EventTupleType, None, None]:
    fromId = None
    while True:
        for i in StatsManager.manager().tailEvents(fromId=fromId):
            yield EventTupleType(
                datetime.datetime.fromtimestamp(i.stamp),
                i.fld1,
                i.fld2,
                i.fld3,
                i.fld4,
                i.event_type,
            )
            fromId = i.pk if i.pk > (fromId or 0) else fromId
        time.sleep(sleepTime)

